from collections import OrderedDict

from transformers import BertTokenizer, is_tf_available
from transformers.modeling_albert import AlbertConfig

from transformers.tokenization_albert import AlbertTokenizer

from ..model_base import TaskModels, ModelType, ModelTaskType, model_func

default_model = model_func(ModelType.albert, AlbertConfig, AlbertTokenizer, 'albert')
bert_model = model_func(ModelType.albert, AlbertConfig, BertTokenizer, 'albert')

'''
Attention please, some albert models are using BertTokenizer.
'''


class Albert_Task_Models_Base(TaskModels):
    MODELS = {
        "cn": [bert_model("clue/albert_chinese_tiny"),
               bert_model("clue/albert_chinese_small"),
               default_model("onePatient/albert_chinese_small"),
               bert_model("voidful/albert_chinese_tiny"),
               bert_model("voidful/albert_chinese_small"),
               bert_model("voidful/albert_chinese_base"),
               bert_model("voidful/albert_chinese_large"),
               bert_model("voidful/albert_chinese_xlarge")
               ]}


class Albert_Task_Models(Albert_Task_Models_Base):
    from transformers.modeling_albert import AlbertForMaskedLM, AlbertModel, AlbertForQuestionAnswering, \
        AlbertForTokenClassification, AlbertPreTrainedModel, AlbertForSequenceClassification
    MODEL_CLASSES = OrderedDict([
        (ModelTaskType.base, AlbertModel),
        (ModelTaskType.pretrain, AlbertPreTrainedModel),
        (ModelTaskType.lm_head, AlbertForMaskedLM),
        (ModelTaskType.seq_cls, AlbertForSequenceClassification),
        (ModelTaskType.token_cls, AlbertForTokenClassification),
        (ModelTaskType.qa, AlbertForQuestionAnswering),
    ])


if is_tf_available():
    class TFAlbert_Task_Models(Albert_Task_Models_Base):
        from transformers.modeling_tf_albert import TFAlbertForMaskedLM, TFAlbertModel, TFAlbertForQuestionAnswering, \
            TFAlbertPreTrainedModel, TFAlbertForSequenceClassification, TFAlbertForMultipleChoice
        MODEL_CLASSES = OrderedDict([
            (ModelTaskType.base, TFAlbertModel),
            (ModelTaskType.pretrain, TFAlbertPreTrainedModel),
            (ModelTaskType.lm_head, TFAlbertForMaskedLM),
            (ModelTaskType.seq_cls, TFAlbertForSequenceClassification),
            (ModelTaskType.multi_choice, TFAlbertForMultipleChoice),
            (ModelTaskType.qa, TFAlbertForQuestionAnswering),
        ])



from torch import nn
import torch
import torch.nn.functional as F


class NewsClassification_WordAVGModel(nn.Module):
    def __init__(self, vocab_size, embedding_dim, output_dim, dropout=0.2, pad_idx=0):
        # 初始化参数，
        super().__init__()
        self.embedding = nn.Embedding(vocab_size, embedding_dim, padding_idx=pad_idx)

        self.fc = nn.Linear(embedding_dim, output_dim)

        self.dropout = nn.Dropout(dropout)

    def forward(self, text):
        # embedded.shape = (batch_size,seq,embed_size)
        embedded = self.dropout(self.embedding(text))

        pooled = F.avg_pool2d(embedded, (embedded.shape[1], 1)).squeeze(1)
        # [batch size, embedding_dim] 把单词长度的维度压扁为1，并降维

        return self.fc(pooled)
        # （batch size,output_dim）


class NewsClassfiction_CNN(nn.Module):
    def __init__(self, vocab_size, embedding_dim, num_filter,
                 filter_sizes, output_dim, dropout=0.2, pad_idx=0):
        super().__init__()

        self.embedding = nn.Embedding(vocab_size, embedding_dim, padding_idx=pad_idx)
        self.convs = nn.ModuleList([
            nn.Conv2d(in_channels=1, out_channels=num_filter,
                      kernel_size=(fs, embedding_dim))
            for fs in filter_sizes
        ])
        # in_channels：输入的channel，文字都是1
        # out_channels：输出的channel维度
        # fs：每次滑动窗口计算用到几个单词,相当于n-gram中的n
        # for fs in filter_sizes用好几个卷积模型最后concate起来看效果。

        self.fc = nn.Linear(len(filter_sizes) * num_filter, output_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, text):
        embedded = self.dropout(self.embedding(text))  # [batch size, sent len, emb dim]
        embedded = embedded.unsqueeze(1)  # [batch size, 1, sent len, emb dim]
        # 升维是为了和nn.Conv2d的输入维度吻合，把channel列升维。
        conved = [F.relu(conv(embedded)).squeeze(3) for conv in self.convs]
        # conved = [batch size, num_filter, sent len - filter_sizes+1]
        # 有几个filter_sizes就有几个conved

        pooled = [F.max_pool1d(conv, conv.shape[2]).squeeze(2) for conv in conved]  # [batch,num_filter]

        cat = self.dropout(torch.cat(pooled, dim=1))
        # cat = [batch size, num_filter * len(filter_sizes)]
        # 把 len(filter_sizes)个卷积模型concate起来传到全连接层。

        return self.fc(cat)


class NewsClassification_RNN(nn.Module):
    def __init__(self, vocab_size, embedding_dim, hidden_dim, output_dim,
                 n_layers=2, bidirectional=True, dropout=0.2, pad_idx=0):
        super().__init__()
        self.embedding = nn.Embedding(vocab_size, embedding_dim, padding_idx=pad_idx)
        self.rnn = nn.LSTM(embedding_dim, hidden_dim, num_layers=n_layers, batch_first=True,
                           bidirectional=bidirectional)

        self.fc = nn.Linear(hidden_dim * 2, output_dim)
        # 这里hidden_dim乘以2是因为是双向，需要拼接两个方向，跟n_layers的层数无关。

        self.dropout = nn.Dropout(dropout)

    def forward(self, text):
        # text.shape=[seq_len, batch_size]
        embedded = self.dropout(self.embedding(text))
        # output: [batch,seq,2*hidden if bidirection else hidden]
        # hidden/cell: [bidirec * n_layers, batch, hidden]
        output, (hidden, cell) = self.rnn(embedded)

        # concat the final forward (hidden[-2,:,:]) and backward (hidden[-1,:,:]) hidden layers
        hidden = self.dropout(torch.cat((hidden[-2, :, :], hidden[-1, :, :]), dim=1))
        # hidden = [batch size, hid dim * num directions]，

        return self.fc(hidden.squeeze(0))  # 在接



import json
import os
import random
from typing import Dict

import numpy as np
import torch
from sklearn.metrics import f1_score
from torch.utils.data import Dataset
from transformers.data.metrics import simple_accuracy

from ..model import ModelMode, task_model
from ..model.model_args import ModelArguments
from ..task.task_args import TaskArguments, parse_tasks_args
from ..train.trainer import Trainer
from ..train.trainer_utils import PredictionOutput, EvalPrediction, InputExample
from ..train.training_args import TrainingArguments
from ..data import DataArguments, TaskDataset, DataProcessor
from ..transformersx_base import aiutils, log, join_path


def set_seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    # ^^ safe to call this function even if cuda is not available


##TODO: need to refactor the TaskModel for various cases
##(1) train a pretrained model from scratch
##(2) finetuning a task based on a pretrained Model
##(3) resume a task from the training interruption
## Currently, if need to resume the training, you must reset the model_path to the checkpoint
class TaskModel:
    def __init__(self, modelArgs: ModelArguments, model_class=None):
        self._model_args = modelArgs
        self._user_model_class = model_class
        self._init()

    def _init(self):
        self.task_model, model_cache_dir = self._build_task_model(self._model_args, self._user_model_class)
        log.info("Loading the task model from {}, {}".format(model_cache_dir, str(self.task_model)))

        self.config, self.tokenizer, self.model = self.task_model.load(
            num_labels=self._model_args.num_labels,
            unit_test=self._model_args.unit_test,
            cache_dir=model_cache_dir
        )

        log.info(
            "Loaded task model, config: {}, tokenizer: {}, "
            "model: {}".format(self.config.to_json_string(False),
                               type(self.tokenizer),
                               type(self.model))
        )

        self._freeze_parameters()

    def _freeze_parameters(self):
        # Currently, freezing the main paramerter is supported only
        self._freeze_weights_main()
        # self._freeze_parameters_others()

        if hasattr(self.model, 'num_parameters'):
            log.info("num params:" + str(self.model.num_parameters()))
            log.info("num trainable params:" + str(self.model.num_parameters(only_trainable=True)))

        if hasattr(self.model, 'named_parameters'):
            log.info("Model Parameters Details: ")
            for name, param in self.model.named_parameters():
                log.info("{}:{}".format(name, param.size()))

    def _freeze_weights_main(self):
        if self.model is None or not self._model_args.freeze_main:
            return

        main_parameters = eval("self.model." + self.task_model.main_parameter)
        if hasattr(main_parameters, "parameters"):
            for param in main_parameters.parameters():
                param.requires_grad = False

    def _freeze_weights_others(self):
        if self.model is None:
            return

        if self._model_args.freeze_parameter and hasattr(self.model, self._model_args.freeze_parameter):
            freeze_model = eval("self.model." + self._model_args.freeze_parameter)
            if hasattr(freeze_model, "parameters"):
                for param in freeze_model.parameters():
                    param.requires_grad = False

    def _build_task_model(self, modelArgs, model_class=None):
        mode_cache_dir = modelArgs.model_pretrained_dir if not modelArgs.not_use_pretrained else modelArgs.model_finetuned_dir

        if not os.path.exists(join_path(mode_cache_dir, modelArgs.model_name)):
            mode_cache_dir = modelArgs.model_pretrained_dir if modelArgs.not_use_pretrained else modelArgs.model_finetuned_dir

        if not os.path.exists(join_path(mode_cache_dir, modelArgs.model_name)):
            raise ValueError(
                "Can not find mode {} in {} or {}".format(modelArgs.model_name, modelArgs.model_finetuned_dir,
                                                          modelArgs.model_pretrained_dir)
            )

        t_model = task_model(model_path=modelArgs.model_name,
                             model_task_type=modelArgs.model_task_type,
                             language=modelArgs.language,
                             framework=modelArgs.framework)

        # use the custom class to replace the model Class
        if model_class is not None:
            t_model.model_class = model_class

        log.info("Built task model: {}".format(str(t_model)))

        return t_model, mode_cache_dir

    def _acc_and_f1(self, preds, labels):
        acc = simple_accuracy(preds, labels)
        f1 = f1_score(y_true=labels, y_pred=preds, average="weighted")
        return {
            "acc": acc,
            "f1": f1,
            "acc_and_f1": (acc + f1) / 2,
        }

    def predict(self, *input):
        return self.model(*input)

    def compute_metrics(self, p: EvalPrediction) -> Dict:
        if self._model_args.model_mode == ModelMode.classification:
            preds = np.argmax(p.predictions, axis=1)
        elif self._model_args.model_mode == ModelMode.regression:
            preds = np.squeeze(p.predictions)
        return self._acc_and_f1(preds, p.label_ids)


class TaskData:
    def __init__(self, dataArgs: DataArguments,
                 tokenizer, dataProcessor: DataProcessor, local_rank):
        self._data_args = dataArgs
        self._tokenizer = tokenizer
        self._processor = dataProcessor
        self._local_rank = local_rank
        self._train_data, self._eval_data = None, None

    def train_data(self):
        if not self._train_data and self._processor:
            self._train_data = TaskDataset(self._data_args,
                                           tokenizer=self._tokenizer,
                                           processor=self._processor,
                                           local_rank=self._local_rank)
        return self._train_data

    def eval_data(self):
        if not self._eval_data and self._processor:
            self._eval_data = TaskDataset(self._data_args,
                                          tokenizer=self._tokenizer,
                                          processor=self._processor,
                                          local_rank=self._local_rank,
                                          evaluate=True)
        return self._eval_data


class TaskTrainer:
    def __init__(self, trainingArgs: TrainingArguments,
                 taskModel: TaskModel, taskData: TaskData, compute_metrics=None):
        self._training_args = trainingArgs
        self._taskModel = taskModel
        self._taskData = taskData
        self._compute_metrics = compute_metrics if compute_metrics else taskModel.compute_metrics
        self._trainer = self._create_trainer()

    def _check_train_args(self):
        if (
                self._training_args.do_train
                and not self._training_args.overwrite_output_dir
                and os.path.exists(self._training_args.output_dir)
                and os.listdir(self._training_args.output_dir)
        ):
            raise ValueError(
                f"Output directory ({self._training_args.output_dir}) already exists and is not empty. Use --overwrite_output_dir to overcome."
            )

    def _create_trainer(self):
        return Trainer(
            model=self._taskModel.model,
            args=self._training_args,
            train_dataset=self._taskData.train_data() if self._training_args.do_train and self._taskData else None,
            eval_dataset=self._taskData.eval_data() if self._training_args.do_eval and self._taskData else None,
            compute_metrics=self._compute_metrics,
        )

    def train(self):
        if self._training_args.do_train:
            # Here, the model_path is for the trainer to load the optimizer and scheduler states
            self._trainer.train(
                model_path=self._taskModel.task_model.model_path
            )
            self._trainer.save_model()
            # For convenience, we also re-save the tokenizer to the same directory,
            # so that you can share your model easily on huggingface.co/models =)
            if self._trainer.is_world_master():
                self._taskModel.tokenizer.save_pretrained(self._training_args.output_dir)

        return self

    def eval(self):
        # Evaluation
        results = {}
        if self._training_args.do_eval and self._training_args.local_rank in [-1, 0]:
            log.info("*** Evaluate ***")

            eval_datasets = [self._taskData.eval_data()]

            for eval_dataset in eval_datasets:
                result = self._trainer.evaluate(eval_dataset=eval_dataset)

                output_eval_file = join_path(
                    self._training_args.output_dir, f"eval_results_{eval_dataset.args.task_name}.txt"
                )
                with open(output_eval_file, "w") as writer:
                    log.info("***** Eval results {} *****".format(eval_dataset.args.task_name))
                    for key, value in result.items():
                        log.info("  %s = %s", key, value)
                        writer.write("%s = %s\n" % (key, value))

                results.update(result)
        return results

    def predict(self, test_dataset: Dataset = None) -> PredictionOutput:
        return self._trainer.predict(test_dataset=self._taskData.eval_data() if not test_dataset else test_dataset)


class TransformerTask:
    def __init__(self, task_name, task_args: TaskArguments,
                 dataProcessor=None, model_class=None, compute_metric=None):
        log.info("Create a TransformerTask with arguments: ")
        log.info(json.dumps(task_args, default=lambda obj: obj.__dict__, indent=True))

        self.task_args = task_args
        self.task_args.validate_and_set_task_name(task_name)
        self._training_args = task_args.training_args

        self._taskModel = TaskModel(task_args.model_args, model_class)
        self._taskData = TaskData(task_args.data_args, self._taskModel.tokenizer, dataProcessor,
                                  task_args.training_args.local_rank) if dataProcessor else None
        self._taskTrainer = TaskTrainer(task_args.training_args, self._taskModel,
                                        self._taskData, compute_metric)

    def _log_task_start(self):
        log.warning(
            "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
            self._training_args.local_rank,
            self._training_args.device,
            self._training_args.n_gpu,
            bool(self._training_args.local_rank != -1),
            self._training_args.fp16,
        )
        log.info("Training/evaluation parameters %s", self._training_args)

    def train(self):
        if not self._training_args.do_train:
            log.warn("do_train argument is set for False")
            return
        self._log_task_start()

        set_seed(self._training_args.seed)
        results = self._taskTrainer.train().eval()
        return results

    def eval(self):
        self._log_task_start()
        set_seed(self._training_args.seed)
        results = self._taskTrainer.eval()
        return results

    def single_predict(self, example: InputExample):
        data_processor = DataProcessor()
        data_processor.get_dev_examples = lambda s: [example]
        return self.predict(data_processor)

    def predict(self, data_processor=None) -> PredictionOutput:
        dataset = TaskDataset(self.task_args.data_args,
                              tokenizer=self._taskModel.tokenizer,
                              processor=data_processor,
                              evaluate=True) if data_processor else None

        return self._taskTrainer.predict(dataset)


class DefaultTask(TransformerTask):
    def __init__(self, dataProcessorClass, model_class=None, compute_metric=None):
        super().__init__("default", parse_tasks_args(), dataProcessorClass, model_class, compute_metric)


def get_transformers_model(models_dir, model_name):
    modelArgs: ModelArguments = ModelArguments()
    modelArgs.model_name = model_name
    modelArgs.model_finetuned_dir = models_dir
    modelArgs.not_use_pretrained = True
    return TaskModel(modelArgs)



from .trainer_base import *
from .trainer_dataloaders import TaskTrainerDataLoaders


class TaskTrainerPredictor(TrainerBase):
    def __init__(self, args: TaskTrainingArguments, model: PreTrainedModel,
                 data_loaders: TaskTrainerDataLoaders):
        super().__init__(args)
        self._model = model
        self._data_loaders = data_loaders

    def predict(self, test_dataset: Dataset) -> TaskPredictionOutput:
        test_dataloader = self._data_loaders.get_test_dataloader(test_dataset)
        return self.prediction_loop(test_dataloader, description="Prediction")

    def _prepare_model(self):
        # multi-gpu eval
        if self._args.n_gpu > 1 and not isinstance(self._model, torch.nn.DataParallel):
            model = torch.nn.DataParallel(self._model)
        else:
            model = self._model
        model.to(self._args.device)
        return model

    def _get_batch_size(self, dataloader: DataLoader):
        if self.is_tpu_available():
            return dataloader._loader._loader.batch_size
        else:
            return dataloader.batch_size

    def prediction_loop(
            self, dataloader: DataLoader, description: str, prediction_loss_only: Optional[bool] = None, max_steps=-1
    ) -> TaskPredictionOutput:
        prediction_loss_only = prediction_loss_only if prediction_loss_only is not None else prediction_loss_only
        model = self._prepare_model()
        batch_size = self._get_batch_size(dataloader)

        logger.info("***** Running %s *****", description)
        logger.info("  Num examples = %d", self.num_examples(dataloader))
        logger.info("  Batch size = %d", batch_size)

        eval_losses: List[float] = []
        preds: np.ndarray = None
        label_ids: np.ndarray = None
        guids = []

        model.eval()

        limit_step = 0
        for inputs in tqdm(dataloader, desc=description):
            if max_steps == 0 or (0 < max_steps < limit_step):
                break
            limit_step = limit_step + 1

            has_labels = any(inputs.get(k) is not None for k in ["labels", "lm_labels", "masked_lm_labels"])

            for k, v in inputs.items():
                if k != 'guid':
                    inputs[k] = v.to(self._args.device)
                else:
                    guids.extend(inputs.pop('guid'))

            with torch.no_grad():
                outputs = model(**inputs)
                if has_labels:
                    step_eval_loss, logits = outputs[:2]
                    eval_losses += [step_eval_loss.mean().item()]
                else:
                    logits = outputs[0]

            if not prediction_loss_only:
                if preds is None:
                    preds = logits.detach().cpu().numpy()
                else:
                    preds = np.append(preds, logits.detach().cpu().numpy(), axis=0)
                if inputs.get("labels") is not None:
                    if label_ids is None:
                        label_ids = inputs["labels"].detach().cpu().numpy()
                    else:
                        label_ids = np.append(label_ids, inputs["labels"].detach().cpu().numpy(), axis=0)

        if self.is_tpu_available() and preds is not None and label_ids is not None:
            # tpu-comment: Get all predictions and labels from all worker shards of eval dataset
            preds = self.xm.mesh_reduce("eval_preds", preds, np.concatenate)
            label_ids = self.xm.mesh_reduce("eval_out_label_ids", label_ids, np.concatenate)

        return TaskPredictionOutput(guids=guids, predictions=preds, label_ids=label_ids, eval_losses=eval_losses)


class TaskTrainerEvaluator(TrainerBase):
    def __init__(self, args: TaskTrainingArguments,
                 predictor: TaskTrainerPredictor,
                 compute_metrics: Optional[Callable[[TaskEvalPrediction], Dict]] = None):
        super().__init__(args)
        self._predictor = predictor
        self._compute_metrics = compute_metrics

    def evaluate(self, dataloader: DataLoader, description: str, prediction_loss_only: Optional[bool] = None,
                 max_steps=-1):
        p = self._predictor.prediction_loop(dataloader, description, prediction_loss_only, max_steps)

        if self._compute_metrics is not None and p.predictions is not None and p.label_ids is not None:
            metrics = self._compute_metrics(TaskEvalPrediction(predictions=p.predictions, label_ids=p.label_ids))
        else:
            metrics = {}
        if len(p.eval_losses) > 0:
            metrics["eval_loss"] = np.mean(p.eval_losses)

        # Prefix all keys with eval_
        for key in list(metrics.keys()):
            if not key.startswith("eval_"):
                metrics[f"eval_{key}"] = metrics.pop(key)
        p.metrics = metrics

        output = json.dumps({**metrics})
        print(output)

        if self._args.tpu_metrics_debug:
            self.xm.master_print(self.met.metrics_report())

        return p

    @staticmethod
    def build(args: TaskTrainingArguments, model: PreTrainedModel,
              data_loaders: TaskTrainerDataLoaders,
              compute_metrics: Optional[Callable[[TaskEvalPrediction], Dict]] = None):
        return TaskTrainerEvaluator(args, TaskTrainerPredictor(args, model, data_loaders), compute_metrics)



from typing import Dict

from transformers.data.metrics import acc_and_f1

from transformers import EvalPrediction

from ai_transformersx.task import TransformerTask
from .news_data_processor import NewsDataProcessor, NewsExampleSegment, PredictDataProcessor, NewsDataArguments
import numpy as np
from ai_transformersx.model import is_turbo_available

from ..task_base import ExampleTaskBase, log, TaskArguments, configclass


@configclass
class NewsSegmentTaskArguments(TaskArguments):
    processor_args: NewsDataArguments = NewsDataArguments()


class NewsSegmentTask(ExampleTaskBase):
    args_class = NewsSegmentTaskArguments

    def __init__(self, taskArgs: NewsSegmentTaskArguments = None, task_class=None):
        super().__init__('news_segment', taskArgs, task_class)
        self._data_processor_args = taskArgs.processor_args

    def _compute_metrics(self, p: EvalPrediction) -> Dict:
        preds = np.argmax(p.predictions, axis=1)
        result = acc_and_f1(preds, p.label_ids)
        log.info(p)
        log.info("preds.size=" + str(len(preds)) +
                 ", preds.sum=" + str(preds.sum()) +
                 ", label.sum=" + str(p.label_ids.sum()))

        correct = ((preds == 0) * (p.label_ids == 0)).sum()
        a1 = correct / (len(p.label_ids) - p.label_ids.sum())

        result['segment_acc'] = a1
        return result

    def _data_processor(self):
        return NewsDataProcessor(self._data_processor_args)

    def predict(self, article: str = None):
        if article:
            segment = NewsExampleSegment(article, context_min_len=self._data_processor_args.context_min_len,
                                         sentence_min_len=self._data_processor_args.sentence_min_len,
                                         check_min_anyway=self._data_processor_args.check_min_anyway)
        else:
            segment = NewsExampleSegment.from_file(join_path(self._data_processor_args.data_dir, 'dev.txt'),
                                                   context_min_len=self._data_processor_args.context_min_len,
                                                   sentence_min_len=self._data_processor_args.sentence_min_len,
                                                   check_min_anyway=self._data_processor_args.check_min_anyway,
                                                   line_sentence=False)
        self.task_args.data_args.predict = True

        p = TransformerTask('news_segment', self.task_args, PredictDataProcessor(segment)).predict()

        preds = np.argmax(p.predictions, axis=1)
        context_indexes = p.guids * (preds == 1)
        separate_indexes = [segment.contexts[i - 1][1] for i in context_indexes if i != 0]

        return segment, separate_indexes


if is_turbo_available():
    from ai_transformersx.model.bert.TurboBertSequenceClassification import TurboBertForSequenceClassification


    class TurboNewsSegmentTask(NewsSegmentTask):
        def __init__(self, taskArgs: NewsSegmentTaskArguments = None):
            super().__init__('turbo_news_segment', taskArgs, TurboBertForSequenceClassification)



import json
import logging
import os
import random
import re
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from pip.utils import cached_property
from torch import nn
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data.sampler import RandomSampler
from tqdm.auto import tqdm, trange
from transformers import is_torch_available

from transformers.data.data_collator import DataCollator, DefaultDataCollator
from transformers.modeling_utils import PreTrainedModel
from transformers.optimization import AdamW, get_linear_schedule_with_warmup
from .trainer_utils import PREFIX_CHECKPOINT_DIR, TaskEvalPrediction, TaskPredictionOutput, TaskTrainOutput, \
    TaskDefaultDataCollatorx
from .training_args import TaskTrainingArguments
from .trainer_utils import *


class TrainerBase:
    def __init__(self, args: TaskTrainingArguments):
        self._args = args

    try:
        import torch_xla.core.xla_model as xm

        _has_tpu = True
    except ImportError:
        _has_tpu = False

    def is_tpu_available(self):
        return self._has_tpu

    try:
        from apex import amp
        _has_apex = True
    except ImportError:
        _has_apex = False

    def is_apex_available(self):
        return self._has_apex

    if _has_tpu:
        import torch_xla.core.xla_model as xm
        import torch_xla.debug.metrics as met
        import torch_xla.distributed.parallel_loader as pl

    def get_tpu_sampler(self, dataset: Dataset):
        if self.xm.xrt_world_size() <= 1:
            return RandomSampler(dataset)
        return DistributedSampler(dataset, num_replicas=self.xm.xrt_world_size(), rank=self.xm.get_ordinal())

    def is_local_master(self) -> bool:
        if self.is_tpu_available():
            return self.xm.is_master_ordinal(local=True)
        else:
            return self._args.local_rank in [-1, 0]

    def is_world_master(self) -> bool:
        """
        This will be True only in one process, even in distributed mode,
        even when training on multiple machines.
        """
        if self.is_tpu_available():
            return self.xm.is_master_ordinal(local=False)
        else:
            return self._args.local_rank == -1 or torch.distributed.get_rank() == 0

    def num_examples(self, dataloader: Union[DataLoader, "pl.PerDeviceLoader"]) -> int:
        """
        Helper to get num of examples from a DataLoader, by accessing its Dataset.
        """
        if self.is_tpu_available():
            assert isinstance(dataloader, self.pl.PerDeviceLoader)
            return len(dataloader._loader._loader.dataset)
        else:
            return len(dataloader.dataset)

    @property
    def train_batch_size(self) -> int:
        return self._args.per_gpu_train_batch_size * max(1, self.n_gpu)

    @property
    def eval_batch_size(self) -> int:
        return self._args.per_gpu_eval_batch_size * max(1, self.n_gpu)

    @cached_property
    def _setup_devices(self) -> Tuple["torch.device", int]:
        logger.info("PyTorch: setting up devices")
        if self._args.no_cuda:
            device = torch.device("cpu")
            n_gpu = 0
        elif self.is_tpu_available():
            device = self.xm.xla_device()
            n_gpu = 0
        elif self._args.local_rank == -1:
            # if n_gpu is > 1 we'll use nn.DataParallel.
            # If you only want to use a specific subset of GPUs use `CUDA_VISIBLE_DEVICES=0`
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            n_gpu = torch.cuda.device_count()
        else:
            # Here, we'll use torch.distributed.
            # Initializes the distributed backend which will take care of sychronizing nodes/GPUs
            torch.distributed.init_process_group(backend="nccl")
            device = torch.device("cuda", self._args.local_rank)
            n_gpu = 1
        return device, n_gpu

    @property
    def device(self) -> "torch.device":
        return self._setup_devices[0]

    @property
    def n_gpu(self):
        return self._setup_devices[1]


def set_seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    # ^^ safe to call this function even if cuda is not available


@contextmanager
def torch_distributed_zero_first(local_rank: int):
    """
    Decorator to make all processes in distributed training wait for the first one (locally) to do something.
    """
    if local_rank not in [-1, 0]:
        torch.distributed.barrier()
    yield
    if local_rank == 0:
        torch.distributed.barrier()


logger = logging.getLogger(__name__)



from .trainer_base import *
from .trainer_counter import TaskTrainedCounter
from .trainer_logger import TaskTrainerLogger
from .trainer_optimizers import TaskTrainerOptimizers
from .trainer_saver import TaskTrainerSaver
from .trainer_dataloaders import TaskTrainerDataLoaders
from .trainer_evaluator import TaskTrainerEvaluator


class TaskTrainerContext(TrainerBase):
    def __init__(self, args: TaskTrainingArguments,
                 model: PreTrainedModel,
                 optimizers: TaskTrainerOptimizers,
                 data_loaders: TaskTrainerDataLoaders,
                 trainer_logger: TaskTrainerLogger,
                 evaluator: TaskTrainerEvaluator):
        super().__init__(args)
        self.model = model
        self.optimizers = optimizers if optimizers else TaskTrainerOptimizers(args, model)
        self.data_loaders = data_loaders
        self.evaluator = evaluator if evaluator else TaskTrainerEvaluator.build(args, model, data_loaders)
        self.trainer_logger = trainer_logger if trainer_logger else TaskTrainerLogger(args, model)
        self.saver = TaskTrainerSaver(args, model)
        self.train_counter = None

        self.__init_env()

    def __init_env(self):
        set_seed(self._args.seed)
        # Create output directory if needed
        if self.is_local_master():
            os.makedirs(self._args.output_dir, exist_ok=True)
        if self.is_tpu_available():
            # Set an xla_device flag on the model's config.
            # We'll find a more elegant and not need to do this in the future.
            self.model.config.xla_device = True

    def init_context_for_train(self, model_path):
        self.train_counter = TaskTrainedCounter(model_path, self._args, len(self.data_loaders.get_train_dataloader()))
        self.saver.load_check_point(model_path, self.optimizers)
        self.optimizers.set_for_start_train()


class TaskTrainStep(TrainerBase):
    def __init__(self, args: TaskTrainingArguments, optimizers: TaskTrainerOptimizers, total_steps):
        self._args = args
        self._model = optimizers.get_model()
        self._optimizer = optimizers.get_optimizer()
        self._scheduler = optimizers.get_scheduler()
        self._total_steps = total_steps

    def _step_backward(self, inputs: Dict[str, torch.Tensor]) -> float:
        self._model.train()
        for k, v in inputs.items():
            inputs[k] = v.to(self._args.device)

        outputs = self._model.forward(**inputs)
        loss = outputs[0]  # model outputs are always tuple in transformers (see doc)

        if self._args.n_gpu > 1:
            loss = loss.mean()  # mean() to average on multi-gpu parallel training
        if self._args.gradient_accumulation_steps > 1:
            loss = loss / self._args.gradient_accumulation_steps

        if self._args.fp16:
            with self.amp.scale_loss(loss, self._optimizer) as scaled_loss:
                scaled_loss.backward()
        else:
            loss.backward()

        return loss.item()

    def _is_step(self, step):
        return (step + 1) % self._args.gradient_accumulation_steps == 0 or (
            # last step in epoch but step is always smaller than gradient_accumulation_steps
                self._total_steps <= self._args.gradient_accumulation_steps
                and (step + 1) == self._total_steps
        )

    def _clip_and_step(self):
        if self._args.fp16:
            torch.nn.utils.clip_grad_norm_(self.amp.master_params(self._optimizer),
                                           self._args.max_grad_norm)
        else:
            torch.nn.utils.clip_grad_norm_(self._model.parameters(), self._args.max_grad_norm)

        if self.is_tpu_available():
            self.xm.optimizer_step(self._optimizer)
        else:
            self._optimizer.step()

        self._scheduler.step()
        self._model.zero_grad()

    def execute(self, step, inputs: Dict[str, torch.Tensor]) -> (float, bool):
        loss = self._step_backward(inputs)

        if not self._is_step(step): return loss, False

        self._clip_and_step()

        return loss, True


class TaskTrainEpoch(TrainerBase):
    def __init__(self, epoch,
                 taskContext: TaskTrainerContext):
        self.task_context = taskContext
        self.epoch_iterator = tqdm(taskContext.data_loaders.get_train_dataloader(), desc="Iteration",
                                   disable=not self.is_local_master())
        self.epoch = epoch
        self.epoch_len = len(self.epoch_iterator)
        self._train_counter = taskContext.train_counter
        self._trainer_step = self.__create_step()

    def __create_step(self):
        return TaskTrainStep(self._args, self.task_context.optimizers, self.epoch_len)

    def execute(self):
        for step, inputs in enumerate(self.epoch_iterator):
            if not self._train_counter.check_step_for_break(self.epoch_iterator):
                break

            if self._train_counter.check_step_for_skip():
                continue

            loss, do_step = self._trainer_step.execute(step, inputs)
            self._process_step(step, do_step, loss)

        self.epoch_iterator.close()
        self.task_context.trainer_logger.log_train_epoch()

    def _process_step(self, step, do_step, loss):
        self._train_counter.on_train_step(step, self.epoch_len, loss)

        if do_step and self.is_local_master():
            self._process_step(self.epoch, self._train_counter.tr_loss,
                               self._train_counter.logging_loss)
            self._train_counter.after_process_step()

    def _process_step_loss(self, epoch, tr_loss, logging_loss):
        if self.task_context.trainer_logger.is_need_log_step(self._train_counter.global_step):
            logs: Dict[str, float] = {}
            logs["loss"] = (tr_loss - logging_loss) / self._train_counter.logging_steps
            logs["learning_rate"] = self.task_context.optimizers.get_scheduler().get_last_lr()[0]

            self.task_context.trainer_logger.log_train_step(epoch, self._train_counter.global_step, logs)

            if self._args.evaluate_during_training:
                self.task_context.evaluator.evaluate(self.task_context.data_loaders.get_eval_dataloader(),
                                                     description="Training setup evaluate, step={}".format(
                                                         self._train_counter.global_step),
                                                     max_steps=self._args.eval_max_steps)

        self.task_context.saver.save_check_point(self.task_context.optimizers, self._train_counter.global_step)


class TaskTrainLoop(TrainerBase):
    def __init__(self, taskContext: TaskTrainerContext):
        self.task_context = taskContext
        self._train_counter = taskContext.train_counter

    def execute(self):
        train_iterator = trange(
            self._train_counter.epochs_trained, int(self._train_counter.num_train_epochs), desc="Epoch",
            disable=not self.is_local_master()
        )

        for epoch in train_iterator:
            if not self._train_counter.check_step_for_break(train_iterator):
                break
            TaskTrainEpoch(epoch, self.task_context).execute()

        train_iterator.close()
        self.task_context.trainer_logger.log_train_end()
        return self._train_counter.task_train_output()


class TaskTrainer():
    def __init__(self, context: TaskTrainerContext):
        self._context = context

    @staticmethod
    def build(args: TaskTrainingArguments,
              model: PreTrainedModel,
              data_loaders: TaskTrainerDataLoaders,
              tb_writer: Optional["SummaryWriter"] = None,
              compute_metrics: Optional[Callable[[TaskEvalPrediction], Dict]] = None):
        TaskTrainer(TaskTrainerContext(args=args,
                                       model=model,
                                       optimizers=TaskTrainerOptimizers(args, model),
                                       data_loaders=data_loaders,
                                       trainer_logger=TaskTrainerLogger(args, model, tb_writer),
                                       evaluator=TaskTrainerEvaluator.build(args, model, data_loaders, compute_metrics)
                                       ))

    def train(self, model_path: Optional[str] = None):
        self._context.init_context_for_train(model_path)

        self._context.trainer_logger.log_pre_train(self._context.data_loaders.get_train_dataloader(),
                                                   self._context.train_counter)

        return TaskTrainLoop(self._context).execute()



import time
from typing import Optional, List

import torch

from ai_transformersx import InputFeatures
from torch.utils.data import Dataset
import os

from ai_transformersx.data.task_data_args import DataArguments
from ..transformersx_base import log, join_path


class TaskDataset(Dataset):
    def __init__(self, features: List[InputFeatures]):
        self.features = features

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i) -> InputFeatures:
        return self.features[i]


class TaskDataStore():
    datastoreId = None
    dataArgs = None

    def load_dataset(self, limit_length: Optional[int] = None, evaluate=False) -> TaskDataset:
        raise NotImplementedError()

    def save_dataset(self, dataset: TaskDataset, evaluate=False):
        raise NotImplementedError()


class LocalDataStore(TaskDataStore):
    def __init__(self, dataStoreId, dataArgs: DataArguments):
        self.datastoreId = dataStoreId
        self.dataArgs = dataArgs

    def load_dataset(self, limit_length: Optional[int] = None, evaluate=False) -> TaskDataset:
        saved_file = self._create_save_file(evaluate)
        if not os.path.exists(saved_file):
            return None
        log.info("Loading dataset from file {}".format(saved_file))
        start = time.time()
        features = torch.load(saved_file)
        log.info(
            f"Loaded dataset from file %s [took %.3f s]", saved_file, time.time() - start
        )
        return TaskDataset(features)

    def _create_save_file(self, evaluate):
        saved_file = "{}_{}".format(self._dataStoreId, ("eval" if evaluate else "train"))
        return join_path(self._dataArgs.data_dir, saved_file)

    def save_dataset(self, dataset: TaskDataset, evaluate=False):
        if not dataset:
            return
        saved_file = self._create_save_file(evaluate)
        if not os.path.exists(saved_file) or self._dataArgs.overwrite:
            log.info("Saving dataset into  file {}".format(saved_file))
            start = time.time()
            torch.save(dataset.features, saved_file)

            log.info(
                f"Saved dataset into cached file %s [took %.3f s]", saved_file, time.time() - start
            )



from .trainer_base import *

from transformersx.train.training_args import TaskTrainingArguments


class TaskTrainerDataLoaders(TrainerBase):
    def __init__(self, args: TaskTrainingArguments,
                 data_collator: DataCollator,
                 train_dataset: Dataset = None,
                 eval_dataset: Dataset = None):
        super().__init__(args)
        self._data_collator = data_collator
        self._train_dataset = train_dataset
        self._eval_dataset = eval_dataset

    def _create_dataloader(self, dataset, sampler, batch_size):
        data_loader = DataLoader(
            dataset,
            sampler=sampler,
            batch_size=batch_size,
            shuffle=False if sampler is not None else True,
            collate_fn=self._data_collator.collate_batch,
        )
        if self.is_tpu_available():
            data_loader = self.pl.ParallelLoader(data_loader, [self._args.device]).per_device_loader(self._args.device)
        return data_loader

    def get_train_dataloader(self) -> DataLoader:
        if self._train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")
        if self.is_tpu_available():
            train_sampler = self.get_tpu_sampler(self._train_dataset)
        else:
            train_sampler = (
                RandomSampler(self._train_dataset)
                if self._args.local_rank == -1
                else DistributedSampler(self._train_dataset)
            )

        return self._create_dataloader(self._train_dataset, train_sampler, self.train_batch_size)

    def get_eval_dataloader(self, eval_dataset: Optional[Dataset] = None) -> DataLoader:
        if eval_dataset is None and self._eval_dataset is None:
            raise ValueError("Trainer: evaluation requires an eval_dataset.")

        eval_dataset = eval_dataset if eval_dataset is not None else self._eval_dataset

        sampler = self.get_tpu_sampler(eval_dataset) if self.is_tpu_available() else None

        return self._create_dataloader(eval_dataset, sampler, self.eval_batch_size)

    def get_test_dataloader(self, test_dataset: Dataset) -> DataLoader:
        # We use the same batch_size as for eval.
        sampler = None
        if self.is_tpu_available():
            sampler = self.get_tpu_sampler(test_dataset) if self.is_tpu_available() else None

        return self._create_dataloader(test_dataset, sampler, self.eval_batch_size)



from ai_harness.fileutils import DirNavigator, join_path
import os
from box import Box
import sys


class ModelFileRname():
    def __init__(self, dir):
        self._dir = dir

    def _handle_file(self, json_file, dir):
        print('Handling ' + join_path(dir, json_file))
        file_base_name = os.path.basename(json_file).replace('.json', '')
        lock_file = join_path(self._dir, dir, file_base_name + '.lock')
        if os.path.exists(lock_file): os.remove(lock_file)

        json_file = join_path(self._dir, dir, json_file)
        obj = Box.from_json(filename=json_file)
        target_name = join_path(self._dir, dir, os.path.basename(obj.url))
        os.rename(join_path(self._dir, dir, file_base_name), target_name)

        os.remove(json_file)

    def run(self):
        dirNavigator = DirNavigator(file_pattern="*.json")
        dirNavigator.on_item(onFile=self._handle_file)
        dirNavigator.nav(self._dir)


if __name__ == '__main__':
    print('Scan dir: ' + sys.argv[1])
    ModelFileRname(sys.argv[1]).run()



from torch.utils.data import Dataset
from transformers import torch_distributed_zero_first
from typing import List, Optional

from .task_data_converter import TaskDataConverter
from .task_data_processor import TaskDataProcessor

from .task_data_store import TaskDataStore, TaskDataset


class TaskDatasetGenerator:
    def __init__(self, data_store: TaskDataStore, processor: TaskDataProcessor,
                 dataConverter: TaskDataConverter):
        self._data_store = data_store
        self._processor = processor
        self._converter = dataConverter

    def dataset(self, predict=False, limit_length: Optional[int] = None, evaluate=False, local_rank=-1) -> Dataset:
        with torch_distributed_zero_first(local_rank):
            dataset = self._data_store.load_dataset(limit_length, evaluate) if not predict else None
            if not dataset:
                features = self.__generate_features(limit_length, evaluate)
                dataset = TaskDataset(features)
                if not predict and local_rank in [-1, 0]:
                    self._data_store.save_dataset(dataset, evaluate)
            return dataset

    def __generate_features(self, limit_length: Optional[int] = None, evaluate=False):
        examples = (
            self._processor.get_eval_examples(limit_length) if evaluate
            else self._processor.get_train_examples(limit_length)
        )

        return self._converter.convert(examples)



from typing import Optional

from transformers import BertConfig, BertTokenizer, TRANSFORMERS_CACHE

from ai_transformersx import *
from ai_transformersx.examples import ExampleManagement, ExampleTaskBase
from ai_transformersx.examples.main import task_manager
from ai_transformersx.model.model_base import _check_and_rename_pretrained_model

import os
from shutil import copytree, rmtree


class TestTransformersxTask:
    task_args: TaskArguments = parse_tasks_args()

    def test_parse_arg(self):
        assert self.task_args.model_args.unit_test == False

    def __do_test_task_model(self, not_use_pretrained):
        self.task_args.model_args.unit_test = True
        self.task_args.model_args.not_use_pretrained = not_use_pretrained
        taskModel: TaskModel = TaskModel(self.task_args.model_args)
        assert isinstance(taskModel.config, (BertConfig))
        assert isinstance(taskModel.tokenizer, (BertTokenizer))
        return taskModel.task_model

    def test_check_and_rename_pretrained_model(self):
        test_path = './models/pretrained/test'
        copytree('./models/pretrained/clue', join_path(test_path, 'clue'))
        _check_and_rename_pretrained_model(test_path, 'clue/albert_chinese_tiny', BertTokenizer)
        rmtree(test_path)

    def test_task_model(self):
        self.task_args.model_args.model_finetuned_dir = "models/finetuning"
        self.task_args.model_args.model_pretrained_dir = "models/pretrained"
        task_model = self.__do_test_task_model(True)
        assert task_model.model_path.startswith("models/finetuning")

        test_path = './models/pretrained/test'
        copytree('./models/pretrained', test_path)
        self.task_args.model_args.model_pretrained_dir = test_path
        task_model = self.__do_test_task_model(False)
        assert not task_model.model_path.startswith("models/finetuning")
        rmtree(test_path)


    def test_parse_task_arguments(self):
        manager = ExampleManagement()
        argument_objs = manager._build_arguments()
        args = ["news", "--action=train", "--processor-args.data-dir=/test/test"]
        task_name, arguments = ComplexArguments(argument_objs).parse(args)
        assert task_name == "news"
        assert arguments.processor_args.data_dir == "/test/test"

    def test_start_example_task(self):
        task_manager.start_example_task(["news", "--action=train", "--processor-args.data-dir=/test/test"], True)



import re

'''
    From open source project HarvestText
'''


def cut_sentences(para, language="cn", drop_empty_line=True, strip=True, deduplicate=False, max_len=-1, min_len=-1):
    '''cut_sentences

    :param para: 输入文本
    :param drop_empty_line: 是否丢弃空行
    :param strip: 是否对每一句话做一次strip
    :param deduplicate: 是否对连续标点去重，帮助对连续标点结尾的句子分句
    :return: sentences: list of str
    '''
    if deduplicate:
        para = re.sub(r"([。！？\!\?])\1+", r"\1", para)

    if language == 'en':
        from nltk import sent_tokenize
        sents = sent_tokenize(para)
        if strip:
            sents = [x.strip() for x in sents]
        if drop_empty_line:
            sents = [x for x in sents if len(x.strip()) > 0]
        return sents
    else:
        para = re.sub('([。！？\?!])([^”’])', r"\1\n\2", para)  # 单字符断句符
        para = re.sub('(\.{6})([^”’])', r"\1\n\2", para)  # 英文省略号
        para = re.sub('(\…{2})([^”’])', r"\1\n\2", para)  # 中文省略号
        para = re.sub('([。！？\?!][”’])([^，。！？\?])', r'\1\n\2', para)
        # 如果双引号前有终止符，那么双引号才是句子的终点，把分句符\n放到双引号后，注意前面的几句都小心保留了双引号
        para = para.rstrip()  # 段尾如果有多余的\n就去掉它
        # 很多规则中会考虑分号;，但是这里我把它忽略不计，破折号、英文双引号等同样忽略，需要的再做些简单调整即可。
        sentences = para.split("\n")

        if strip:
            sentences = [sent.strip() for sent in sentences]
        if drop_empty_line:
            sentences = [sent for sent in sentences if len(sent.strip()) > 0]

        if max_len<1:
            return sentences

        new_sentences = []
        for s in sentences:
            news = []
            while len(s) > max_len + min_len:
                news.append(s[:max_len])
                s = s[max_len:]
            if news:
                new_sentences.extend(news)
            else:
                new_sentences.append(s)
        return new_sentences



from .trainer_base import *
from .training_args import TaskTrainingArguments
from .trainer_optimizers import TaskTrainerOptimizers


class TaskTrainerSaver(TrainerBase):
    def __init__(self, args: TaskTrainingArguments,
                 model: PreTrainedModel):
        super().__init__(args)
        self._model = model

    def save_check_point(self, trainer_optimizers: TaskTrainerOptimizers, global_step):
        if self._args.save_steps > 0 and global_step % self._args.save_steps == 0:
            # In all cases (even distributed/parallel), self.model is always a reference
            # to the model we want to save.
            model = trainer_optimizers.get_model()
            if hasattr(model, "module"):
                assert model.module is self._model
            else:
                assert model is self._model
            # Save model checkpoint
            output_dir = os.path.join(
                self._args.output_dir, f"{PREFIX_CHECKPOINT_DIR}-{global_step}"
            )

            self._save_check_point_model(output_dir)
            self._rotate_checkpoints()

            torch.save(trainer_optimizers.get_optimizer().state_dict(), os.path.join(output_dir, "optimizer.pt"))
            torch.save(trainer_optimizers.get_scheduler().state_dict(), os.path.join(output_dir, "scheduler.pt"))
            logger.info("Saving optimizer and scheduler states to %s", output_dir)

    def load_check_point(self, model_path, trainer_optimizers: TaskTrainerOptimizers):
        if (
                model_path is not None
                and os.path.isfile(os.path.join(model_path, "optimizer.pt"))
                and os.path.isfile(os.path.join(model_path, "scheduler.pt"))
        ):
            # Load in optimizer and scheduler states
            logger.info("Loading optimizer and scheduler from :" + model_path)
            trainer_optimizers.get_optimizer().load_state_dict(torch.load(os.path.join(model_path, "optimizer.pt")))
            trainer_optimizers.get_scheduler().load_state_dict(torch.load(os.path.join(model_path, "scheduler.pt")))

    def _sorted_checkpoints(self, checkpoint_prefix=PREFIX_CHECKPOINT_DIR, use_mtime=False) -> List[str]:
        ordering_and_checkpoint_path = []

        glob_checkpoints = [str(x) for x in Path(self._args.output_dir).glob(f"{checkpoint_prefix}-*")]

        for path in glob_checkpoints:
            if use_mtime:
                ordering_and_checkpoint_path.append((os.path.getmtime(path), path))
            else:
                regex_match = re.match(f".*{checkpoint_prefix}-([0-9]+)", path)
                if regex_match and regex_match.groups():
                    ordering_and_checkpoint_path.append((int(regex_match.groups()[0]), path))

        checkpoints_sorted = sorted(ordering_and_checkpoint_path)
        checkpoints_sorted = [checkpoint[1] for checkpoint in checkpoints_sorted]
        return checkpoints_sorted

    def _rotate_checkpoints(self, use_mtime=False) -> None:
        if self._args.save_total_limit is None or self._args.save_total_limit <= 0:
            return

        # Check if we should delete older checkpoint(s)
        checkpoints_sorted = self._sorted_checkpoints(use_mtime=use_mtime)
        if len(checkpoints_sorted) <= self._args.save_total_limit:
            return

        number_of_checkpoints_to_delete = max(0, len(checkpoints_sorted) - self._args.save_total_limit)
        checkpoints_to_be_deleted = checkpoints_sorted[:number_of_checkpoints_to_delete]
        for checkpoint in checkpoints_to_be_deleted:
            logger.info("Deleting older checkpoint [{}] due to args.save_total_limit".format(checkpoint))
            shutil.rmtree(checkpoint)

    def _save_check_point_model(self, output_dir: Optional[str] = None):
        """
        Saving best-practices: if you use default names for the model,
        you can reload it using from_pretrained().

        Will only save from the master process.
        """

        if self.is_tpu_available():
            self._save_check_point_tpu(output_dir)

        elif self.is_world_master():
            self._save_check_point(output_dir)

    def _save_check_point_tpu(self, output_dir: Optional[str] = None):
        output_dir = output_dir if output_dir is not None else self._args.output_dir
        logger.info("Saving model checkpoint to %s", output_dir)

        if self.xm.is_master_ordinal():
            os.makedirs(output_dir, exist_ok=True)
            torch.save(self._args, os.path.join(output_dir, "training_args.bin"))

        # Save a trained model and configuration using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        if not isinstance(self._model, PreTrainedModel):
            raise ValueError("Trainer.model appears to not be a PreTrainedModel")

        self.xm.rendezvous("saving_checkpoint")
        self._model.save_pretrained(output_dir)

    def _save_check_point(self, output_dir: Optional[str] = None):
        output_dir = output_dir if output_dir is not None else self._args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info("Saving model checkpoint to %s", output_dir)
        # Save a trained model and configuration using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        if not isinstance(self._model, PreTrainedModel):
            raise ValueError("Trainer.model appears to not be a PreTrainedModel")
        self._model.save_pretrained(output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(self._args, os.path.join(output_dir, "training_args.bin"))



from collections import OrderedDict

from transformers import is_tf_available
from transformers.modeling_bert import BertConfig
from transformers.tokenization_bert import BertTokenizer, BertTokenizerFast, BertWordPieceTokenizer

from ..model_base import TaskModels, ModelType, ModelTaskType, model_func

default_model = model_func(ModelType.bert, BertConfig, BertTokenizer, 'bert')
fast_model = model_func(ModelType.bert, BertConfig, BertTokenizerFast, 'bert')
wpe_model = model_func(ModelType.bert, BertConfig, BertWordPieceTokenizer, 'bert')


class Bert_Task_Models_Base(TaskModels):
    MODELS = {
        "cn": [default_model("bert-base-chinese"),
               default_model("adamlin/bert-distil-chinese"),
               default_model("hfl/chinese-bert-wwm"),
               default_model("hfl/chinese-bert-wwm-ext")
               ]}


class Bert_Task_Models(Bert_Task_Models_Base):
    from transformers.modeling_bert import BertModel, BertForPreTraining, BertForMaskedLM, \
        BertForMultipleChoice, BertForNextSentencePrediction, BertForTokenClassification, BertForSequenceClassification
    MODEL_CLASSES = OrderedDict([
        (ModelTaskType.base, BertModel),
        (ModelTaskType.pretrain, BertForPreTraining),
        (ModelTaskType.lm_head, BertForMaskedLM),
        (ModelTaskType.seq_cls, BertForSequenceClassification),
        (ModelTaskType.multi_choice, BertForMultipleChoice),
        (ModelTaskType.token_cls, BertForTokenClassification),
        (ModelTaskType.next_seq, BertForNextSentencePrediction),
    ])


if is_tf_available():
    class TFBert_Task_Models(Bert_Task_Models_Base):
        from transformers.modeling_tf_bert import TFBertModel, TFBertForPreTraining, TFBertForMaskedLM, \
            TFBertForMultipleChoice, TFBertForNextSentencePrediction, TFBertForTokenClassification, \
            TFBertForSequenceClassification
        MODEL_CLASSES = OrderedDict([
            (ModelTaskType.base, TFBertModel),
            (ModelTaskType.pretrain, TFBertForPreTraining),
            (ModelTaskType.lm_head, TFBertForMaskedLM),
            (ModelTaskType.seq_cls, TFBertForSequenceClassification),
            (ModelTaskType.multi_choice, TFBertForMultipleChoice),
            (ModelTaskType.token_cls, TFBertForTokenClassification),
            (ModelTaskType.next_seq, TFBertForNextSentencePrediction),
        ])



from .data_args import TaskDataArguments
from .data_processor import TaskDataProcessor
from .data_store import TaskDataset, LocalDataStore
from .dataset_factory import TaskDatasetFactory
from .data_converter import TaskDataConverter
from .data_models import TaskInputExample, TaskInputFeatures



import torch
from fast_bert import accuracy_thresh, roc_auc, fbeta, accuracy

from .fast_task_args import FastTaskArguments
from fast_bert.learner_abs import BertAbsDataBunch, BertAbsLearner
from fast_bert.learner_cls import BertDataBunch, BertLearner
from fast_bert.learner_lm import BertLMDataBunch, BertLMLearner
from fast_bert.learner_qa import BertQADataBunch, BertQALearner
from fast_bert.prediction import BertClassificationPredictor
from ..transformersx_base import *


class TransformersFastTask():
    def __init__(self, task_name, args: FastTaskArguments):
        self._args = args
        self._task_name = task_name


class TransformersFastAbsTask(TransformersFastTask):
    def __init__(self, task_name, args: FastTaskArguments):
        super().__init__(task_name, args)


class TransformersFastClsTask(TransformersFastTask):
    def __init__(self, task_name, args: FastTaskArguments):
        super().__init__(task_name, args)
        self._model_path = join_path(self._args.model_pretrained_dir, self._args.model_name)
        self._output_dir = join_path(self._args.model_finetuned_dir, task_name, self._args.model_name)

    def metrics(self):
        metrics = []
        metrics.append({'name': 'accuracy', 'function': accuracy})
        # metrics.append({'name': 'roc_auc', 'function': roc_auc})
        # metrics.append({'name': 'fbeta', 'function': fbeta})
        return metrics

    def train(self):
        databunch = BertDataBunch(self._args.data_dir, self._args.data_dir,
                                  tokenizer=self._model_path,
                                  train_file=self._args.train_file,
                                  val_file=self._args.eval_file,
                                  label_file=self._args.labels_file,
                                  text_col=self._args.text_col,
                                  label_col=self._args.label_col,
                                  batch_size_per_gpu=self._args.batch_size_per_gpu,
                                  max_seq_length=self._args.max_seq_length,
                                  multi_gpu=True,
                                  multi_label=self._args.multi_label,
                                  model_type='bert')
        device = torch.device('cuda')
        learner = BertLearner.from_pretrained_model(databunch, self._model_path, metrics=self.metrics(),
                                                    device=device, logger=log, output_dir=self._output_dir,
                                                    finetuned_wgts_path=None, warmup_steps=5,
                                                    multi_gpu=True, is_fp16=self._args.fp16,
                                                    multi_label=self._args.multi_label, logging_steps=0)
        learner.fit(self._args.num_train_epochs, self._args.learning_rate, validate=True)
        learner.validate()


class TransformersFastLmTask(TransformersFastTask):
    def __init__(self, task_name, args: FastTaskArguments):
        super().__init__(task_name, args)


class TransformersFastQaTask(TransformersFastTask):
    def __init__(self, task_name, args: FastTaskArguments):
        super().__init__(task_name, args)



from .trainer import TaskTrainer, TaskTrainLoop, TaskTrainedCounter, TaskTrainEpoch, TaskTrainStep, TaskTrainerContext, \
    TaskTrainerOptimizers, TaskTrainerSaver, TaskTrainerEvaluator, TaskTrainerLogger, TaskTrainerDataLoaders, \
    TaskTrainingArguments, TaskEvalPrediction, TaskPredictionOutput, TaskTrainOutput, TaskDefaultDataCollatorx
